#include <stdarg.h>
#include <cstring>

#include "midi.hh"

void MIDItrack::Add(unsigned char data)
{
    track.push_back(data);
}
void MIDItrack::Flush()
{
    AddVarLen(delay); delay = 0;
}
void MIDItrack::AddVarLen(unsigned t)
{
    //fprintf(stderr, "%p:Appending varlen %u\n", this,t);
    if(t>>21)Add(0x80 | ((t>>21) &127));
    if(t>>14)Add(0x80 | ((t>>14) &127));
    if(t>>7) Add(0x80 | ((t>>7)  &127));
    Add(t&127);
}
void MIDItrack::AddText(unsigned char n, const char* t)
{
    Flush();
    Add(0xFF);
    Add(n);
    Add(strlen(t));
    while(*t)Add(*t++);
}
void MIDItrack::AddData(unsigned char data)
{
    if(data & 0x80)
    {
        Flush();
        if(data == running)return;
        running = data;
    }
    Add(data);
}
void MIDItrack::AddData(unsigned char nbytes, unsigned char data, ...)
{
    va_list ap;
    va_start(ap, data);
    for(;;)
    {
    	AddData(data);
    	if(nbytes <= 1)break;
    	--nbytes;
    	data = (unsigned char)(va_arg(ap, int));
    }
    va_end(ap);
}
void MIDItrack::AddMeta(unsigned char metatype, unsigned char nbytes, ...)
{
	Flush();
    Add(0xFF);
    Add(metatype);
    Add(nbytes);
    va_list ap;
    va_start(ap, nbytes);
    while(nbytes > 0)
    {
        unsigned char byte = (unsigned char)(va_arg(ap, int));
        Add(byte);
        --nbytes;
    }
    va_end(ap);
}
unsigned MIDItrack::ReadVarLen()
{
    unsigned result = 0;
    while(readpos < track.size())
    {
        unsigned char c = track[readpos++];
        result = result*128 + (c & 127);
        if(c < 128) break;
    }
    return result;
}
void MIDItrack::ReadDelay()
{
    if(readpos < track.size()) delay = ReadVarLen();
}
int MIDItrack::ReadMeta(std::string& data)
{
    if(readpos+3 <= track.size()
    && track[readpos] == 0xFF)
    {
        unsigned char type = track[readpos+1];
        readpos += 2;
        unsigned length = ReadVarLen();
        unsigned orig = readpos;
        readpos += length;
        data = std::string(&track[orig], &track[orig+length]);
        return type;
    }
    return -1;
}
void MIDItrack::PeekNextCmd(std::string& cmd, bool peek_only)
{
    cmd = "";
    unsigned save_pos = readpos;
    try
    {
        unsigned char b = track.at(readpos);
        
        if(b < 0x80)
            b = running;
        else
        {
            ++readpos;
            if(b < 0xF0) running = b;
        }
        
        unsigned origpos = readpos;
        
        if(b == 0xFF)
        {
            unsigned char type = track.at(readpos++);
            unsigned p = ReadVarLen();
            readpos += p;
        }
        else if(b == 0xF0 || b == 0xF7)
        {
            unsigned p = ReadVarLen();
            readpos += p;
        }
        else if(b == 0xF2) readpos += 2;
        else if(b == 0xF3) readpos += 1;
        else if(b >= 0xC0 && b <= 0xCF) readpos += 1;
        else if(b >= 0x80 && b <= 0xEF) readpos += 2;
        
        if(b & 0x80)
            cmd += (char)b;
        while(origpos < readpos)
            cmd += (char)(track.at(origpos++));
    }
    catch(...)
    {
        delay=1;
    }
    if(peek_only) readpos = save_pos;
}
void MIDItrack::ReadEventInto(MIDIfile& target, unsigned trackno)
{
    std::string cmd;
    PeekNextCmd(cmd, false);
    
    if(!cmd.empty())
    {
        /*
        fprintf(stderr, "Append: ");
        for(unsigned a=0; a<cmd.size(); ++a)
            fprintf(stderr, " %02X", (unsigned char)cmd[a]);
        fprintf(stderr, " (delay=%u)\n", target.tracks[trackno].GetDelay());
        */
        
        if(((unsigned char)cmd[0]) >= 0xF0)
        {
            target.tracks[trackno].Flush();
            
            for(unsigned a=0; a<cmd.size(); ++a)
                target.tracks[trackno].Add(cmd[a]);
        }
        else
        {
            target.tracks[trackno].AddData(cmd[0]);
            for(unsigned a=1; a<cmd.size(); ++a)
                target.tracks[trackno].Add(cmd[a]);
        }
    }
}

void MIDIfile::InitTrack(unsigned trackno)
{
    if(tracks.size() <= trackno)
    {
        tracks.reserve(16);
        tracks.resize(trackno+1);
    }
    
    tracks[trackno].Reset();
    
    // Meta 0x58 (misc settings):
        //      time signature: 4/2
        //      ticks/metro:    24
        //      32nd per 1/4:   8
    tracks[trackno].AddMeta(0x58, 4, 4,2, 24,8);
    
    // Meta 0x51 (tempo):
    tracks[trackno].AddMeta(0x51, 3, ((tempo>>16)&255),
                                     ((tempo>>8)&255),
                                     ((tempo)&255));
}
void MIDIfile::Done()
{
    AddLoopEnd();
    for(unsigned a=0; a<tracks.size(); ++a)
    {
        // Meta 0x2F (track end marker):
        tracks[a].AddMeta(0x2F, 0);
    }
    
    midi.clear();
    
    // Estimate the size of the MIDI file.
    unsigned midisize=10+2+2;
    for(unsigned a=0; a<tracks.size(); ++a)
        midisize += tracks[a].size() + 8;
    midi.reserve(midisize);
    
    for(unsigned c=0; c<10; ++c)
        midi.push_back(
            "MThd\0\0\0\6" // MIDI signature (MThd and number 6)
            "\0\1"         // Format 1 (multiple tracks, synchoronous)
              [c]);
    
    // Track count
    midi.push_back((tracks.size() >> 8) & 255);
    midi.push_back((tracks.size()     ) & 255);
    
    midi.push_back((deltaticks>>8)&255);
    midi.push_back((deltaticks   )&255);
    
    for(unsigned a=0; a<tracks.size(); ++a)
    {
        const MIDItrack& trk = tracks[a];
        
        for(unsigned c=0; c<4; ++c)
            midi.push_back("MTrk"[c]);
        
        const unsigned tracklen = trk.size();
        
        midi.push_back((tracklen>>24)&255);
        midi.push_back((tracklen>>16)&255);
        midi.push_back((tracklen>> 8)&255); 
        midi.push_back((tracklen    )&255);
        
        midi.insert(midi.end(), trk.getdata(), trk.getdata() + tracklen);
    }
}
bool MIDIfile::Load(const char* filename)
{
    unsigned char header[14];
    FILE* fp = fopen(filename, "rb");
    if(!fp) return false;
    
    fread(header,1,sizeof(header), fp);
    if(memcmp(header, "MThd\0\0\0\6", 8) != 0)
    {
    euh:
        fclose(fp);
        return false;
    }
    unsigned ntracks = header[10]*256 + header[11];
    deltaticks = header[12]*256 + header[13];
    for(unsigned a=0; a<ntracks; ++a)
    {
        InitTrack(a);
        fread(header,1,8,fp);
        if(memcmp(header,"MTrk",4) != 0) goto euh;
        unsigned tracklen = header[4];
        tracklen = tracklen*256+header[5];
        tracklen = tracklen*256+header[6];
        tracklen = tracklen*256+header[7];
        
        std::vector<unsigned char> buf(tracklen);
        fread(&buf[0], 1, tracklen, fp);
        
        tracks[a].SetData(&buf[0], tracklen);
    }
    return true;
}

void MIDIfile::KeyOn(unsigned trackno, unsigned ch, unsigned char note, unsigned char pressure)
{
    if(false&&trackno == 0)
    {
        fprintf(stderr, "Track %u: delay(%d) noteon[%u] %02X %02X\n",
            trackno, tracks[trackno].GetDelay(), ch, note, pressure);
    }
    tracks[trackno].AddData(3, 0x90|ch, note, pressure);
}

void MIDIfile::KeyOff(unsigned trackno, unsigned ch, unsigned char note, unsigned char pressure)
{
    if(false&&trackno == 0)
    {
        fprintf(stderr, "Track %u: delay(%d) noteoff[%u] %02X %02X\n",
            trackno, tracks[trackno].GetDelay(), ch, note, pressure);
    }
    tracks[trackno].AddData(3, 0x80|ch, note, pressure);
}

void MIDIfile::KeyTouch(unsigned trackno, unsigned ch, unsigned char note, unsigned char pressure)
{
    if(false&&trackno == 0)
    {
        fprintf(stderr, "Track %u: delay(%d) touch[%u] %02X %02X\n",
            trackno, tracks[trackno].GetDelay(), ch, note, pressure);
    }
    tracks[trackno].AddData(3, 0xA0|ch, note, pressure);
}

void MIDIfile::Control(unsigned trackno, unsigned ch, unsigned char ctrl, unsigned char value)
{
    if(false&&trackno == 0)
    {
        fprintf(stderr, "Track %u: delay(%d) ontrol[%u] %02X %02X\n",
            trackno, tracks[trackno].GetDelay(), ch, ctrl, value);
    }
    tracks[trackno].AddData(3, 0xB0|ch, ctrl, value);
}

void MIDIfile::Patch(unsigned trackno, unsigned ch, unsigned patchno)
{
    if(false&&trackno == 0)
    {
        fprintf(stderr, "Track %u: delay(%d) patch[%u] %02X\n",
            trackno, tracks[trackno].GetDelay(), ch, patchno);
    }
    tracks[trackno].AddData(2, 0xC0|ch, patchno);
}

void MIDIfile::Wheel(unsigned trackno, unsigned ch, unsigned value)
{
    tracks[trackno].AddData(3, 0xE0|ch, value&127, (value>>7)&127);
}

void MIDIfile::Save(const char* filename)
{
    FILE* fp = fopen(filename, "wb");
    if(!fp) return;
    
    fwrite(getdata(), 1, size(), fp);
    fclose(fp);
}
